// -*- mode: c++ -*-
// $Id: Mesh.cpp,v 1.6 2004/01/31 03:19:19 ps Exp $
// $Source: /cs/research/multires/cvsroot/src/frame/Mesh.cpp,v $

#include <iostream>
#include <map>
#include <algorithm>
#include <vector>
#include <set>

#include "Mesh.h"
#include "Vertex.h"
#include "Triangle.h"

Mesh::~Mesh( void )
{
  // be a good citizen and clean up after yourself...
  for_each( VertexBegin(), VertexEnd(), Delete<Vertex>() );
  for_each( TriangleBegin(), TriangleEnd(), Delete<Triangle>() );
}

inline int
Token( std::istream& in, char* str, int len, const char* delim )
{
  // initialize to empty, i.e., put a zero in the leading entry
  *str=0;
  const char* pos;

  // one at a time...
  char c = in.get();

  //extract leading delims
  while( in.good() && strchr( delim, c ) ) c = in.get();

  while( len > 0 && in.good() ){
    if( ( pos = strchr( delim, c ) ) ) return int( *pos );
    *(str++) = c;
    *str = 0;
    len--;
    c = in.get();
    if( !in ) return -1;
  }
  return -1;
}

inline void
skipEOL( std::istream& in ){
  while( true ){
    char c=in.get();
    if( !in || c == '\n' ) return;
  }
}

std::istream&
operator>>( std::istream& is, Mesh& m )
{
  // read in vert coords and triangles(!)
  const int BUFSIZE = 256;
  static char
    szBuf[BUFSIZE],
    szDummy[BUFSIZE],
    sz1[BUFSIZE],
    sz2[BUFSIZE],
    sz3[BUFSIZE],
    delims[4]={ '\n', ' ', '\t', 0 };
  // we do this in chunks
  const int CHUNKS = 40;
  static char szn[CHUNKS][BUFSIZE];

  // read so far
  int nVerts = 0, nTris = 0;

  while( is.good() ){

    int res = Token( is, szBuf, BUFSIZE, delims); 
    if( res == -1 || res == '\n' ) continue;

    // all the things we don't care about
    if( szBuf[0] == '#' ||
	szBuf[0] == 'g' ||
	szBuf[0] == 'l' ||
	szBuf[0] == 's'){
      skipEOL(is);
      continue;
    }

    int nitems = 0;
    if( szBuf[0] == 'f' ){
      while( res != '\n' && res != -1 && nitems < CHUNKS ){
        res = Token( is, szn[nitems], BUFSIZE, delims ); 
        ++nitems;
      }
    }else{
      assert( szBuf[0] == 'v' );
      // if the first element is not 'f' we assume(!) that we are
      // reading 3 floats here...
      res = Token( is, sz1, BUFSIZE, delims ); 
      if( res == -1 || res == '\n' ) continue;

      res = Token( is, sz2, BUFSIZE, delims ); 
      if( res == -1 || res == '\n' ) continue;

      res = Token( is, sz3, BUFSIZE, delims ); 
      if( res == -1 ) continue;

      // PS: this should move outside this if/else; haven't tested
      // this enough though.
      // then discard the rest of the line
      while( res != '\n' ){
	if( ( res = Token( is, szDummy, BUFSIZE, delims ) ) == -1 ){
	  break;      
	}
      }
    }

    if( szBuf[0] == 'v' ){
      Vertex *v = new Vertex; assert( v );
      v->p() = CVec( float( atof( sz1 ) ),
		     float( atof( sz2 ) ),
		     float( atof( sz3 ) ) );
      m.addVertex( v );
      ++nVerts;
    }else if( szBuf[0] == 'f' ){
      // we are only handling triangles...
      assert( nitems == 3 );
      // .obj files use 1 offset!
      Triangle *t = new Triangle( m.m_vc[atoi(szn[0])-1],
				  m.m_vc[atoi(szn[1])-1],
				  m.m_vc[atoi(szn[2])-1] );
      assert( t );
      // get anything the Triangle may want (currently empty)
      m.addTriangle( t );
      ++nTris;
    }else{ // can't determine what the line means just skip it then
      skipEOL(is);
      continue;
    }
  }

  // all the initialization code associated with a fully loaded Mesh
  m.Init();

  return is;
}

std::ostream&
operator<<( std::ostream& os, const Mesh& m )
{
  // something we don't really need to do, but will in preparation for
  // sparse matrix structures: build all edges as they correspond to
  // non-zero entries of the matrix and then sort the rows and write
  // things out in that order; first we need to know how many edges
  // there are; It's the sum total of all valencies
  Mesh::const_VertexIt vi = m.VertexBegin(), ve = m.VertexEnd();

  int i = 0;

  os << "# obj file\n# v: "
     << (unsigned int)m.VertexSize() << " f: "
     << (unsigned int)m.TriangleSize() << " e: "
    // this the total number of off-diagonal entries in the matrix
	 << i << std::endl;

  std::map<Vertex*,int> vmap;

  os.precision(12);
  vi = m.VertexBegin();
  for( i = 0; vi != ve; ++vi, ++i ){
    // the vertex print function prints the 3 floats
	if( !( os << "v " << *(*vi) << std::endl ) ) return os;
    // keep track of address to index mapping
    vmap[*vi] = i;
  }
  Mesh::const_TriangleIt ti = m.TriangleBegin(), te = m.TriangleEnd();
  for( ; ti != te; ++ti ){
    const Triangle *t = *ti;
    // obj files are 1-offset
    if( !( os << "f "
	   << vmap[t->a()]+1 << " "
	   << vmap[t->b()]+1 << " "
	   << vmap[t->c()]+1 << std::endl ) ) return os;
  }

  return os;
}

